using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Diagnostics;
using System.Threading;
using System.Text.RegularExpressions;

namespace WindowFocusHook
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender,EventArgs e)
        {
            SetSplash();
            InstallHook();
        }

        private void InstallHook()
        {
            // Create an instance of HookProc.
            MouseHookProcedure = new HookProc(Form1.MouseHookProc);

            /* 
             * Some notes about how to configure this project.
             * 
             * - SetWindowsHookEx fails if the Visual Studio Hosting
             *   Process is enabled; so disable it in project
             *   properties. See a similar CodeProject app:
             *   http://www.codeproject.com/script/Articles/ViewDownloads.aspx?aid=7294
             *   
             * - Also, you need to build as Release, not Debug.
             */
            hHook = SetWindowsHookEx(
                        WH_MOUSE_LL,
                        MouseHookProcedure,
                        Marshal.GetHINSTANCE(
                            Assembly.GetExecutingAssembly().GetModules()[0]),
                        0);
            //If the SetWindowsHookEx function fails.
            if(hHook == 0)
            {
                /* Maybe check Marshal.GetLastWin32Error here someday.
                 * At least some guys on the net seem to do it. */
                MessageBox.Show(this,"SetWindowsHookEx Failed",
                    "Windows Hook Error");
                return;
            }
        }

        private void Unhook()
        {
            bool ret = UnhookWindowsHookEx(hHook);
            //If the UnhookWindowsHookEx function fails.
            if(ret == false)
            {
                MessageBox.Show(this,"UnhookWindowsHookEx Failed",
                    "Windows Hook Error");
                return;
            }
            hHook = 0;
        }

        /* See https://msdn.microsoft.com/en-us/library/windows/desktop/ms644986(v=vs.85).aspx */
        public delegate int HookProc(int nCode,IntPtr wParam,IntPtr lParam);

        //Declare the hook handle as an int.
        private static int hHook = 0;

        //Declare the mouse hook constant.
        //For other hook types, you can obtain these values from Winuser.h in the Microsoft SDK.
        public const int WH_MOUSE_LL = 14;

        private enum MouseMessages
        {
            WM_LBUTTONDOWN = 0x0201,
            WM_LBUTTONUP = 0x0202,
            WM_MOUSEMOVE = 0x0200,
            WM_RBUTTONDOWN = 0x0204,
            WM_RBUTTONUP = 0x0205,
        }

        [Flags()]
        private enum SetWindowPosFlags : uint
        {
            None = 0x0000,
            AsynchronousWindowPosition = 0x4000,
            DeferErase = 0x2000,
            DrawFrame = 0x0020,
            FrameChanged = 0x0020,
            HideWindow = 0x0080,
            DoNotActivate = 0x0010,
            DoNotCopyBits = 0x0100,
            IgnoreMove = 0x0002,
            DoNotChangeOwnerZOrder = 0x0200,
            DoNotRedraw = 0x0008,
            DoNotReposition = 0x0200,
            DoNotSendChangingEvent = 0x0400,
            IgnoreResize = 0x0001,
            IgnoreZOrder = 0x0004,
            ShowWindow = 0x0040,
        }

        public enum FocusApi
        {
            SetForegroundWindow,
            BringWindowToTop,
            SetWindowPos,
            SetFocus,
        }

        //Declare MouseHookProcedure as a HookProc type.
        private HookProc MouseHookProcedure;

        /* See https://msdn.microsoft.com/en-us/library/windows/desktop/ms644970(v=vs.85).aspx */
        [StructLayout(LayoutKind.Sequential)]
        public class MouseHookStruct
        {
            public Point pt;
            public int mouseData;
            public int flags;
            public int time;
            public int hwnd;
            public int dwExtraInfo;
        }

        //This is the Import for the SetWindowsHookEx function.
        [DllImport("user32.dll")]
        private static extern int SetWindowsHookEx(int idHook,HookProc lpfn,
            IntPtr hInstance,int threadId);

        //This is the Import for the UnhookWindowsHookEx function.
        //Call this function to uninstall the hook.
        [DllImport("user32.dll")]
        private  static extern bool UnhookWindowsHookEx(int idHook);

        [DllImport("user32.dll")]
        private static extern IntPtr WindowFromPoint(Point point);

        [DllImport("user32.dll",SetLastError = true)]
        private static extern int GetClassName(IntPtr hWnd, StringBuilder
            lpString,int nMaxCount);

        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool SetForegroundWindow(IntPtr hWnd);

        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool BringWindowToTop(IntPtr hWnd);

        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter,
            int X, int Y, int cx, int cy, SetWindowPosFlags uFlags);

        [DllImport("user32.dll")]
        private static extern IntPtr GetAncestor(IntPtr hWnd, uint gaFlags);
        
        [DllImport("user32.dll", SetLastError = true)]
        private static extern uint GetWindowThreadProcessId(IntPtr hWnd,
            out uint lpdwProcessId);

        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool AttachThreadInput(int idAttach, int idAttachTo,
            bool fAttach);

        [DllImport("user32.dll")]
        private static extern IntPtr SetFocus(IntPtr hWnd);

        [DllImport("user32.dll")]
        private static extern IntPtr GetForegroundWindow();

        [DllImport("kernel32.dll")]
        private static extern uint GetCurrentThreadId();
        
        //This is the Import for the CallNextHookEx function.
        //Use this function to pass the hook information to the next hook procedure in chain.
        [DllImport("user32.dll")]
        private static extern int CallNextHookEx(int idHook, int nCode,
            IntPtr wParam,IntPtr lParam);

        private static FocusApi focusApi = FocusApi.SetForegroundWindow;
        private static bool attachThread = false;
        private static bool debug = false;
        private static string infoStr = "";
        private static bool useTopLevelWindow = true;
        private static IntPtr lastHwnd = IntPtr.Zero;
        private static string pattern = ".*";
        private static Regex patternRegex = new Regex(".*");
        private static int windowEnterTime = 0;
        private static IntPtr windowEnterHandle = IntPtr.Zero;
        private static int delay = 4; /* in 100 ms */
        private static int tickCount = 0;
        private static StringBuilder windowClassNameBuf = new StringBuilder(50);
        private static bool rbuttonDown = false;
        private static bool lbuttonDown = false;
        /* Getting the middle button isn't as straight-forward, so we don't
         * bother. Is mostly mouse-wheel-emulation stuff on Windows, should
         * be rare enough. */

        private int hideGreeterCount = 0;
        private int hideGreeterDelay = 80;

        public static void SetFocusApi(FocusApi focusApi)
        {
            Trace.TraceInformation("focusApi=" + focusApi);
            Form1.focusApi = focusApi;
        }
        public static void SetAttachThread(bool attachThread)
        {
            Trace.TraceInformation("attachThread="+attachThread);
            Form1.attachThread = attachThread;
        }
        public static void SetDebug(bool debug)
        {
            Trace.TraceInformation("debug=" + debug);
            Form1.debug = debug;
        }
        public static void SetUseTopLevelWindow(bool useTopLevelWindow)
        {
            Trace.TraceInformation("useTopLevelWindow=" + useTopLevelWindow);
            Form1.useTopLevelWindow = useTopLevelWindow;
        }
        public static void SetDelay(int delay)
        {
            Trace.TraceInformation("delay=" + delay);
            Form1.delay = delay;
        }
        public static void SetPattern(string pattern)
        {
            Trace.TraceInformation("pattern=" + pattern);
            Form1.pattern = pattern;
            patternRegex = new Regex(pattern);
        }

        private static string GetWindowClassName(IntPtr hwnd)
        {
            /* Store in a static buffer as side-effect. Not really
             * beautiful but it should save a memory allocation
             * in many cases, as ToString() might copy the string
             * even if the StringBuilder was freshly allocated. */
            GetClassName(hwnd, windowClassNameBuf, windowClassNameBuf.Capacity);
            string windowClsStr = windowClassNameBuf.ToString();
            return windowClsStr;
        }
        
        public static int MouseHookProc(int nCode,IntPtr wParam,IntPtr lParam)
        {
            //Marshall the data from the callback.
            MouseHookStruct MyMouseHookStruct = 
                (MouseHookStruct)Marshal.PtrToStructure(lParam,typeof(MouseHookStruct));

            if(nCode < 0)
            {
                return CallNextHookEx(hHook,nCode,wParam,lParam);
            }
            else
            {
                switch ((MouseMessages)wParam)
                {
                    case MouseMessages.WM_LBUTTONDOWN:
                        lbuttonDown = true;
                        break;
                    case MouseMessages.WM_RBUTTONDOWN:
                        rbuttonDown = true;
                        break;
                    case MouseMessages.WM_LBUTTONUP:
                        lbuttonDown = false;
                        break;
                    case MouseMessages.WM_RBUTTONUP:
                        rbuttonDown = false;
                        break;
                    case MouseMessages.WM_MOUSEMOVE:
                        IntPtr hwnd;
                        Point point = MyMouseHookStruct.pt;
                        hwnd = WindowFromPoint(point);
                        if (useTopLevelWindow)
                        {
                            IntPtr hwnd2 = GetAncestor(hwnd, 2);
                            hwnd = hwnd2;
                        }
                        if (debug) infoStr = "x = " +
                            point.X.ToString("d") +
                            " y = " +
                            point.Y.ToString("d") +
                            " lbutton = " + lbuttonDown +
                            " rbutton = " + rbuttonDown +
                            " mbutton = " + rbuttonDown +
                            " hwnd = " + hwnd +
                            " time = " + MyMouseHookStruct.time +
                            " cls = " + GetWindowClassName(hwnd);
                        if (!lastHwnd.Equals(hwnd))
                        {
                            string windowClsStr = GetWindowClassName(hwnd);
                            if (patternRegex.IsMatch(windowClsStr))
                            {
                                windowEnterHandle = hwnd;
                                windowEnterTime = MyMouseHookStruct.time;
                                tickCount = 0;
                                if (debug) Trace.TraceInformation("Setting windowEnterTime to " +
                                     windowEnterTime+"; windowEnterHandle="+windowEnterHandle);
                            }
                            else
                            {
                                windowEnterTime = 0;
                            }
                        }
                        lastHwnd = hwnd;
                        break;
                }
                return CallNextHookEx(hHook, nCode, wParam, lParam);
            }
        }

        private void SetSplash()
        {
            textBox1.Lines = new string[]{
                "-= Window Focus Hook =-",
                "",
                "Raises windows according to a class name pattern.",
                "",
                "This program will run indefinitely once started."+
                "Use 'taskkill /im:WindowFocusHook.exe' to stop it.",
                "",
                "This message will disappear in a few seconds."
            };
            textBox1.Select(0, 0);
        }

        private static void ActivateWin()
        {
            bool result=false;
            uint thisThreadId=0;
            uint activePid=0,activeThreadId = 0;
            uint enterPid=0,enterThreadId=0;
            
            if(attachThread)
            {
                thisThreadId = GetCurrentThreadId();
                IntPtr activeWindow = GetForegroundWindow();
                activeThreadId = GetWindowThreadProcessId(activeWindow,
                    out activePid);
                enterThreadId = GetWindowThreadProcessId(windowEnterHandle,
                    out enterPid);
                if(thisThreadId != activeThreadId)
                {
                    if(debug) Trace.TraceInformation("Attaching thread "+
                        thisThreadId + " to " + activeThreadId);
                    result = AttachThreadInput((int)thisThreadId,
                        (int)activeThreadId,true);
                    if (!result) Trace.TraceError("AttachThreadInput failed");
                }
                if(enterThreadId != thisThreadId)
                {
                    if(debug) Trace.TraceInformation("Attaching thread "+
                        enterThreadId + " to " + thisThreadId);
                    result = AttachThreadInput((int)enterThreadId,
                        (int)thisThreadId,true);
                    if (!result) Trace.TraceError("AttachThreadInput failed");
                }
            }
            
            switch(focusApi)
            {
                case FocusApi.SetWindowPos:
                    result = SetWindowPos(windowEnterHandle, IntPtr.Zero, 0, 0, 0, 0,
                        SetWindowPosFlags.AsynchronousWindowPosition |
                        SetWindowPosFlags.IgnoreMove |
                        SetWindowPosFlags.IgnoreResize);
                    if(debug) Trace.TraceInformation("SetWindowPos("+
                        windowEnterHandle+",...)");
                    if (!result) Trace.TraceError("SetWindowPos failed");
                    break;
                    
                case FocusApi.SetFocus:
                    if(debug) Trace.TraceInformation("SetFocus("+
                        windowEnterHandle+")");
                    SetFocus(windowEnterHandle);
                    break;
                    
                case FocusApi.SetForegroundWindow:
                    if(debug) Trace.TraceInformation("SetForegroundWindow("+
                        windowEnterHandle+")");
                    result = SetForegroundWindow(windowEnterHandle);
                    if (!result) Trace.TraceError("SetForegroundWindow failed");
                    break;
                    
                case FocusApi.BringWindowToTop:
                    if(debug) Trace.TraceInformation("BringWindowToTop("+
                        windowEnterHandle+")");
                    result = BringWindowToTop(windowEnterHandle);
                    if (!result) Trace.TraceError("BringWindowToTop failed");
                    break;
            }
            
            if(attachThread)
            {
                if (enterThreadId != thisThreadId)
                {
                    if(debug) Trace.TraceInformation("Unattaching thread "+
                        enterThreadId + " from " + thisThreadId);
                    result = AttachThreadInput((int)enterThreadId,
                        (int)thisThreadId, true);
                    if (!result) Trace.TraceError("AttachThreadInput failed");
                }
                if (thisThreadId != activeThreadId)
                {
                    if(debug) Trace.TraceInformation("Unattaching thread "+
                        thisThreadId + " from " + activeThreadId);
                    result = AttachThreadInput((int)thisThreadId,
                        (int)activeThreadId, true);
                    if (!result) Trace.TraceError("AttachThreadInput failed");
                }
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if(debug) textBox1.Text = infoStr;
            if(hideGreeterCount<hideGreeterDelay)
            {
                hideGreeterCount++;
            }
            else
            {
                if (!debug && this.Visible) this.Visible = false;
            }
            if(windowEnterTime!=0)
            {
                if (tickCount >= delay)
                {
                    if (debug) Trace.TraceInformation("Activating window");
                    ActivateWin();
                    windowEnterTime = 0;
                }
                else
                {
                    if(!(lbuttonDown || rbuttonDown)) tickCount++;
                }
            }
        }

    }
}