using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Diagnostics;

/*
 * If you ever should need to compile this program without msbuild.exe,
 * use:
 *
 * csc /out:WindowFocusHook.exe /target:winexe Form1.cs Form1.Designer.cs Program.cs
 *
 * To enable trace output, use this instead:
 *
 * csc /out:WindowFocusHook.exe /d:TRACE Form1.cs Form1.Designer.cs Program.cs
 */
namespace WindowFocusHook
{
    static class Program
    {

        private static void ShowUsage(string error)
        {
            string usageText =
                "Incorrect parameters specified.\n"+
                "\n"+
                "Required command line arguments:\n"+
                "\n"+
                "-raisedelay <delay-in-100-ms> (raise matched windows after delay)\n"+
                "\n"+
                "Optional command line arguments:\n"+
                "\n"+
                "-debug (enables debug mode)\n"+
                "-classpattern <regex-pattern> (pattern to match window class name)\n"+
                "-actonchild (makes pattern match child windows, not top-level)\n"+
                "-focusapi SetFocus|BringWindowToTop|SetForegroundWindow|SetWindowPos\n"+
                "    (sets API to use on matched window; default: SetForegroundWindow)\n"+
                "-attachthread (attaches program to matched window's input thread)\n"+
                "\n"+
                "Error was:\n"+
                error;
            MessageBox.Show(usageText,"Usage Error");
        }

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static int Main(string[] args)
        {
            bool raise=false;
            int delay=4;
            bool actOnTopLevel=true;
            string classpattern = ".*";
            bool debug = false;
            bool attachThread = false;
            Form1.FocusApi focusApi = Form1.FocusApi.SetForegroundWindow;

            try
            {
                for (int optind = 0; optind < args.Length; optind++)
                {
                    string arg = args[optind];
                    switch (arg)
                    {
                        case "-raisedelay":
                            delay = Convert.ToInt32(args[++optind]);
                            raise = true;
                            break;
                        case "-classpattern":
                            classpattern = args[++optind];
                            break;
                        case "-debug":
                            ConsoleTraceListener listener = 
                                new ConsoleTraceListener();
                            Trace.Listeners.Add(listener);
                            debug = true;
                            break;
                        case "-actonchild":
                            actOnTopLevel = false;
                            break;
                        case "-attachthread":
                            attachThread = true;
                            break;
                        case "-focusapi":
                            focusApi = (Form1.FocusApi)Enum.Parse(typeof(Form1.FocusApi),
                                args[++optind]);
                            break;
                        default:
                            ShowUsage("Unknown argument " + arg);
                            return 1;
                    }
                }
            }
            catch(Exception e)
            {
                ShowUsage(e.Message);
                return 1;
            }

            if(!raise)
            {
                ShowUsage("-raisedelay is required");
                return 1;
            }

            Form1.SetFocusApi(focusApi);
            Form1.SetAttachThread(attachThread);
            Form1.SetDelay(delay);
            Form1.SetPattern(classpattern);
            Form1.SetUseTopLevelWindow(actOnTopLevel);
            Form1.SetDebug(debug);

            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new Form1());

            return 0;
        }
    }
}